/**************************************************************************
*  Copyright (c) 2015 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  17.10.2015  mifi  First Version, based on my Xflash application:
*                    http://www.emb4fun.de/archive/xflash/index.html
*  07.01.2016  mifi  Use adler32 instead of crc32_ether.
*  15.08.2020  mifi  Added ZLIB support
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <windows.h>
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>

#include "xfile.h"
#include "adler32.h"
#include "zlib.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define MAX_FILES       1024
#define XFILE_NAME      "xfile.bin"

#define VERSION         "1.08"
#define MAX_INPUT_SIZE  (4*1024*1024)

#define MIN(a, b) ((a) < (b) ? (a) : (b))

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static char szInFolder[_MAX_PATH];     /* Input folder name */
static DWORD dSectorSize = 0;          /* Sector size CRC32 */
static DWORD dProductID = 0;           /* Product ID */
static DWORD dAlignment = 0;           /* Alignment */
static char szConfigFile[_MAX_PATH];   /* Configuration file e.g. etc\config.txt */

static BYTE   Image[MAX_INPUT_SIZE];
static BYTE   ImageComp[MAX_INPUT_SIZE];

static char  *Filelist[MAX_FILES];
static int   nFilelistCounter = 0;
static int   nUseZLIB = 0;

static char szDataName[20] = {0};
static DWORD dDataVersion  = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  OutputStartMessage                                                   */
/*                                                                       */
/*  Output start message.                                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputStartMessage (void)
{
   printf("\n");   
   printf("xfile v%s compiled "__DATE__" "__TIME__"\n", VERSION);
   printf("(c) 2015 by Michael Fischer (www.emb4fun.de)\n");
   printf("\n");
} /* OutputStartMessage */

/*************************************************************************/
/*  OutputUsage                                                          */
/*                                                                       */
/*  Output "usage" message.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputUsage (void)
{
  printf("Usage: xfile -i:InFolder [-s:CRC32Size] [-p:ProductID] [-c:Config] [-z]\n");
  printf("                         [-a:Alignment]\n");
  printf("\n");
  printf("  -i   Input folder, e.g. -i:html\n");
  printf("  -s   Size for sectored CRC32, e.g. -s:512\n");
  printf("       Default 0, to disable this functionality\n");
  printf("  -p   Product ID (hex), e.g. -p:0x80000001\n");
  printf("  -c   Config file, e.g. -c:etc\\config.txt\n");
  printf("  -z   Use ZLIB for compression\n");
  printf("  -a   Size alignment, e.g. -a:128\n");
  
} /* OutputUsage */

/*************************************************************************/
/*  Cleanup                                                              */
/*                                                                       */
/*  Free the Filelist if available                                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void Cleanup (int rc)
{
   WORD wIndex;

   for(wIndex=0; wIndex<MAX_FILES; wIndex++)
   {
      if (Filelist[wIndex] != NULL)
      {
         free(Filelist[wIndex]);
      }
   }      
   
   exit(rc);
} /* Cleanup */

/*************************************************************************/
/*  Add2Filelist                                                         */
/*                                                                       */
/*  Add the given filename to the list.                                  */
/*                                                                       */
/*  In    : pStartPath                                                   */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void Add2Filelist (char *pFilename)
{
  int nLen;

  if ((pFilename != NULL) && (nFilelistCounter < MAX_FILES)) {
    nLen = strlen(pFilename);
    Filelist[nFilelistCounter] = (char*)malloc(nLen+1);
    memset(Filelist[nFilelistCounter], 0x00, nLen+1);
    memcpy(Filelist[nFilelistCounter], pFilename, nLen);
    nFilelistCounter++;
  } else {
    if (nFilelistCounter >= MAX_FILES) {
      printf("Error: To many files > %d\n", MAX_FILES);
      Cleanup(1);
    }
  }
} /* Add2Filelist */

/*************************************************************************/
/*  FindAllFiles                                                         */
/*                                                                       */
/*  Find all files by the given start folder.                            */
/*                                                                       */
/*  In    : pStartPath                                                   */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void FindAllFiles (char *pStartPath)
{
  HANDLE           hFile;
  WIN32_FIND_DATA   FindData;
  char            szSearchPath[_MAX_PATH];
  char            szFilename[_MAX_PATH];
  
  memset(&FindData, 0, sizeof(FindData));
  sprintf(szSearchPath, "%s\\*.*", pStartPath);

  hFile = FindFirstFile(szSearchPath, &FindData);
  if (hFile == INVALID_HANDLE_VALUE)
    return;
  do {
    if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
      if (FindData.cFileName[0] != '.') {
        sprintf(szSearchPath, "%s\\%s", pStartPath, FindData.cFileName);
        FindAllFiles(szSearchPath);
      }
    } else {
      sprintf(szFilename, "%s\\%s", pStartPath, FindData.cFileName); 
      Add2Filelist(szFilename);
    }
  }
  while (FindNextFile(hFile, &FindData) != 0);

  FindClose(hFile);
} /* FindAllFiles */

/*************************************************************************/
/*  WriteImage                                                           */
/*                                                                       */
/*  Write the image file to the disk.                                    */
/*                                                                       */
/*  In    : pHeader, dHeaderSize, pImage, dImageSize                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WriteImage (XFILE_HEADER *pHeader, DWORD dHeaderSize, BYTE *pImage, DWORD dImageSize)
{
   FILE *hOutFile;
   DWORD dWriteCnt;
   DWORD dSectorCount;
   DWORD dRestCount;
   DWORD dExpectedSize;
   DWORD dIndex;
   DWORD dCRC32;
   DWORD dAlignmentBytes = 0;
   BYTE  bAB = 0xFF; /* Alignment Byte */
   int   nRet;
   DWORD dImageCompSize;   


   /*
    * Check if data must be compressed
    */
   if (1 == nUseZLIB)
   {
      dImageCompSize = sizeof(ImageComp);
      nRet = compress(ImageComp, (uLongf*)&dImageCompSize, pImage, dImageSize);
      if (nRet != 0)
      {
         /* ZLIB error */
         printf("Error: ZLIB error\n");
         Cleanup(1);
      }
      pHeader->dDataCompSize  = dImageCompSize;
      pHeader->dDataCompCRC32 = adler32(ADLER_START_VALUE, ImageComp, dImageCompSize);
      
      /* Create new header CRC */
      pHeader->dHeaderCRC32   = adler32(ADLER_START_VALUE, (char*)pHeader, sizeof(XFILE_HEADER) - XFILE_SIZE_OF_CRC32);
      
      pImage     = ImageComp;
      dImageSize = dImageCompSize; 
   }    
   

   /*
    * Create output file
    */
   hOutFile = fopen(XFILE_NAME, "wb");
   if (hOutFile != NULL)
   {
      /*
       * Check if sector CRC is selected 
       */
      if (0 == dSectorSize)
      {
         /* 
          * No sector CRC is selected, create output file 
          */
         /* Write Image to the disk */
         dWriteCnt  = fwrite(pHeader, sizeof(BYTE), dHeaderSize, hOutFile);
         dWriteCnt += fwrite(pImage, sizeof(BYTE), dImageSize, hOutFile);
         
         /* Check if an aligment is needed */   
         if (dAlignment != 0)
         {
            /* Calculate aligment bytes */
            dAlignmentBytes = dAlignment - ((dHeaderSize + dImageSize) % dAlignment);
            
            /* Write alignment bytes */
            for(dIndex=0; dIndex<dAlignmentBytes; dIndex++)
            {
               dWriteCnt += fwrite(&bAB, sizeof(BYTE), 1, hOutFile);
            }
         }
         
         fclose(hOutFile);

         /* Check data count written to the disk */
         if (dWriteCnt == (dHeaderSize + dImageSize + dAlignmentBytes))
         {
            printf("Output file \"%s\" was successful created, %d bytes\n", XFILE_NAME, (dHeaderSize + dImageSize + dAlignmentBytes));        
         }
         else
         {
            /* Write error */
            printf("Error: Write error\n");
            DeleteFile(XFILE_NAME);
            Cleanup(1);
         }
      }
      else
      {
         /* 
          * Sector CRC is selected
          */
         dSectorCount = dImageSize / dSectorSize;
         dRestCount   = dImageSize % dSectorSize;
      
         /* Calculate expected size */
         dExpectedSize = dHeaderSize;
      
         /* Write header */      
         dWriteCnt = fwrite(pHeader, sizeof(BYTE), dHeaderSize, hOutFile);
      
         /* Write dSectorCount * (sector data + CRC) */
         if (dSectorCount != 0)
         {
            dExpectedSize += dSectorCount * (dSectorSize + XFILE_SIZE_OF_CRC32);

            for(dIndex=0; dIndex<dSectorCount; dIndex++)
            {
               dCRC32     = adler32(ADLER_START_VALUE, pImage, dSectorSize);
               dWriteCnt += fwrite(pImage, sizeof(BYTE), dSectorSize, hOutFile);
               dWriteCnt += fwrite(&dCRC32, sizeof(BYTE), XFILE_SIZE_OF_CRC32, hOutFile);
         
               pImage += dSectorSize;
            }
         }         

         /* Write dRestCount + CRC */
         if (dRestCount != 0)
         {
            dExpectedSize += dRestCount + XFILE_SIZE_OF_CRC32;

            dCRC32     = adler32(ADLER_START_VALUE, pImage, dRestCount);
            dWriteCnt += fwrite(pImage, sizeof(BYTE), dRestCount, hOutFile);
            dWriteCnt += fwrite(&dCRC32, sizeof(BYTE), XFILE_SIZE_OF_CRC32, hOutFile);
         }         
      
         /* Close output file */
         fclose(hOutFile);

         /* Check data count written to the disk */
         if (dWriteCnt != dExpectedSize)
         {
            /* Write error */
            printf("Error: Write error\n");
            DeleteFile(XFILE_NAME);
            Cleanup(1);
         }
      }
   }
   else
   {
      printf("Error: Could not create \"%s\"\n", XFILE_NAME);
      Cleanup(1);
   }

} /* WriteImage */ 

/*************************************************************************/
/*  CreateOutputImage                                                    */
/*                                                                       */
/*  Create the output image file.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void CreateOutputImage (void)
{
   BYTE            *pBuffer;
   BYTE            *pXFileStart;
   XFILE_FAT_ENTRY *pFATStart;
   XFILE_FAT_ENTRY *pFAT;
   BYTE            *pNameStart;
   BYTE            *pDataStart;
   DWORD            dImageSize;
   WORD             wIndex;
   LONG             lInFolderLen;
   char            *pFilename;
   long             lNamelength;
   int              nFile;
   LONG             lFilelength;   
   XFILE_HEADER      Header;
   LONG             lUnixtime;
   char            *pChar;
   size_t            Len;
   
   
   /*
    * The output image will look like:
    *
    *                 -----------
    *                | Header    |
    *                 -----------
    *    xfile start | Entry #1  |  "FAT" section     
    *                | Entry #2  |
    *                |     .     |
    *                |     .     |
    *                | Entry #n  |
    *                | Entry end | 
    *                 -----------
    *                | Name #1   |  Name section
    *                | Name #2   |
    *                |     .     |
    *                |     .     |
    *                | Name #n   |
    *                 -----------
    *                | Data #1   |  Data section
    *                | Data #2   |
    *                |     .     |
    *                |     .     |
    *                | Data #n   |
    *                 -----------
    */

   /* "Clear" Image */
   memset(Image, 0xFF, sizeof(Image));
   
   pBuffer    = &Image[0];
   dImageSize = 0;
   
   pXFileStart = (BYTE*)pBuffer;
   pFATStart   = (XFILE_FAT_ENTRY*)pBuffer;
   pFAT        = (XFILE_FAT_ENTRY*)pBuffer;

   /* 
    * Create "FAT" section 
    */
   pBuffer    += (nFilelistCounter+1) * sizeof(XFILE_FAT_ENTRY);
   dImageSize += (nFilelistCounter+1) * sizeof(XFILE_FAT_ENTRY);
   
   /*
    * Create "Name" section
    */
   pNameStart   = pBuffer;
   lInFolderLen = strlen(szInFolder);
   
   for(wIndex=0; wIndex<nFilelistCounter; wIndex++)
   {
      pFilename   = Filelist[wIndex];
      pFilename  += lInFolderLen + 1;              /* Jump over "InFolder\" */
      lNamelength = strlen(pFilename) + 1;         /* Add 1 of the 0 */
      lNamelength = ((lNamelength + 3) / 4) * 4;   /* Aligned to 4 */
      
      memset(pBuffer, 0x00, lNamelength);
      strcpy(pBuffer, pFilename);
      
      /* Create offset to pXFileStart */ 
      pFAT[wIndex].dFilename = ((DWORD)pBuffer - (DWORD)pXFileStart);
      
      pBuffer    += lNamelength; 
      dImageSize += lNamelength;
   }
   
   /* Set FAT end marker */
   pFAT[wIndex].dFilename   = 0;
   pFAT[wIndex].dFilelength = 0;
   pFAT[wIndex].dData       = 0;
   
   /*
    * Create "Data" section
    */
   pDataStart = pBuffer;
       
   for(wIndex=0; wIndex<nFilelistCounter; wIndex++)
   {
      nFile = _open(Filelist[wIndex], _O_BINARY | _O_RDONLY);
      if (nFile != -1) 
      {
         /* Get file length */
         lFilelength = _filelength(nFile);
         pFAT[wIndex].dFilelength = lFilelength;
         
         if ((dImageSize + lFilelength) < MAX_INPUT_SIZE)
         {
            /* Read data */   
            _read(nFile, pBuffer, lFilelength);
            _close(nFile);
         
            /* Create offset to pXFileStart */ 
            pFAT[wIndex].dData = ((DWORD)pBuffer - (DWORD)pXFileStart);

            /* Aligned to 4 */ 
            lFilelength = ((lFilelength + 3) / 4) * 4;
         
            pBuffer    += lFilelength; 
            dImageSize += lFilelength;
         }
         else
         {
            _close(nFile);
            printf("Error: xfile size > %d\n", MAX_INPUT_SIZE);
            Cleanup(1);
         }            
      }
   }
   
   if ((sizeof(XFILE_HEADER) + dImageSize) < MAX_INPUT_SIZE)
   {
      /*
       * Before we can create the xfile.bin, the file name must be 
       * converter for the html access. Change "\" to "/" and
       * use lower chars.
       */
      for(wIndex=0; wIndex<nFilelistCounter; wIndex++)
      {
         pChar = (char*)((DWORD)pXFileStart + pFAT[wIndex].dFilename); 
         while (*pChar != 0)
         {
            switch (*pChar)
            {
               case '\\':
               {
                  *pChar = '/';
                  break;
               }
               
               case '0':
               case '1':
               case '2':
               case '3':
               case '4':
               case '5':
               case '6':
               case '7':
               case '8':
               case '9':
               case '_':
               {
                  /* Do nothing */
                  break;
               }
               
               default:
               {
                  if (isupper(*pChar))
                  {
                     *pChar = _tolower(*pChar);
                  }
                  break;
               }
            }
            pChar++;
         }
      }
       
      /* Get Unixtime */
      _tzset();
      time(&lUnixtime);
   
      /*
       * Create XFILE header
       */
      memset(&Header, 0x00, sizeof(XFILE_HEADER));
   
      Header.dMagic1           = XFILE_HEADER_MAGIC_1;
      Header.dMagic2           = XFILE_HEADER_MAGIC_2;
      Header.dSizeVersion      = XFILE_HEADER_SIZEVER;
      Header.dProductID        = dProductID;
      Header.dDataTotalSize    = dImageSize;
      Header.dDataSectorSize   = dSectorSize;
      Header.dDataCreationTime = lUnixtime;
      Header.dDataCRC32        = adler32(ADLER_START_VALUE, Image, dImageSize);
      
      Len = strlen(szDataName);
      Len = MIN(Len, XFILE_HEADER_NAME_SIZE-1);
      memcpy(Header.DataName, szDataName, Len); 
      Header.dDataVersion      = dDataVersion;
      
      Header.dHeaderCRC32      = adler32(ADLER_START_VALUE, (char*)&Header, sizeof(XFILE_HEADER) - XFILE_SIZE_OF_CRC32);
      
      WriteImage(&Header, sizeof(XFILE_HEADER), Image, dImageSize);       
   }      
   else
   {
      printf("Error: xfile size > %d\n", MAX_INPUT_SIZE);
      Cleanup(1);
   }            

} /* CreateOutputImage */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : argc, argv                                                   */
/*  Out   : none                                                         */
/*  Return: 0 / error cause                                              */
/*************************************************************************/
int main (int argc, char **argv)
{
   int   nIndex;
   int   nLen;
   char *pString;
   char szIni[_MAX_PATH];
   UINT   Digit1, Digit2, Digit3;
   UINT   Digit4, Digit5, Digit6;

   szInFolder[0]   = 0;
   dSectorSize     = 0;
   szConfigFile[0] = 0;
   
   memset(Filelist, 0x00, sizeof(Filelist));
   
   OutputStartMessage();
   
   if (argc < 2)
   {
      OutputUsage();
   }
   else
   {
      /* 
       * Parse command line 
       */
      for (nIndex=0; nIndex<argc; nIndex++)
      {
         /* Input folder */
         if ((pString = strstr(argv[nIndex], "-i:")) != NULL)
         {
            pString += 3;  /* Jump over "-i:" */
            nLen     = strlen(pString);
            if (nLen < (sizeof(szInFolder)-1))
            {
               strcpy(szInFolder, pString);
            }
            else
            {
               printf("Error: Input folder name len > %d\n", (sizeof(szInFolder)-1));
               Cleanup(1);
            }
         } /* Parse input file name */
         
         /* SectorSize */
         if ((pString = strstr(argv[nIndex], "-s:")) != NULL)
         {
            pString += 3;  /* Jump over "-s:" */
            dSectorSize = atoi(pString);
            
            if ((dSectorSize % 4) != 0)
            {
               printf("Error: SectorSize must be aligned to 4\n");
               Cleanup(1);
            }
         } /* Parse SectorSize */
         
         /* ProductID */
         if ((pString = strstr(argv[nIndex], "-p:")) != NULL)
         {
            pString += 3;  /* Jump over "-p:" */
            
            sscanf(pString, "%x", &dProductID);
         } /* Parse DestAddr */
         
         /* Config file */
         if ((pString = strstr(argv[nIndex], "-c:")) != NULL)
         {
            pString += 3;  /* Jump over "-c:" */
            nLen     = strlen(pString);
            if (nLen < (sizeof(szConfigFile)-1))
            {
               strcpy(szConfigFile, pString);
            }
            else
            {
               printf("Error: Config filer name len > %d\n", (sizeof(szConfigFile)-1));
               Cleanup(1);
            }
         } /* Parse config file name */
         
         /* Use ZLIB */
         if ((pString = strstr(argv[nIndex], "-z")) != NULL)
         {
            nUseZLIB = 1;
         } /* Parse ZLIB */

         /* Alignment */
         if ((pString = strstr(argv[nIndex], "-a:")) != NULL)
         {
            pString += 3;  /* Jump over "-a:" */
            dAlignment = atoi(pString);
         } /* Parse Alignment */
         
      } /* end for loop */
      
      if (0 == szInFolder[0])
      {
         printf("Error: No input folder specified\n");
         Cleanup(1);
      }
      
      /*
       * Here the command line was parsed and no error was available
       */
      printf("Input folder: %s\n", szInFolder);
      printf("CRC size    : %d\n", dSectorSize);
      printf("Product ID  : %08X\n", dProductID);
      printf("Alignment   : %d\n", dAlignment);
      printf("\n");

      /* Get info like name and version */
      if (szConfigFile[0] != 0)
      {  
         sprintf(szIni, "%s\\%s", szInFolder, szConfigFile);
         GetPrivateProfileString("system", "name", "unknown", szDataName, sizeof(szDataName), szIni); 
         Digit1 = GetPrivateProfileInt("version", "digit1", 0, szIni);
         Digit2 = GetPrivateProfileInt("version", "digit2", 0, szIni);
         Digit3 = GetPrivateProfileInt("version", "digit3", 0, szIni);
         Digit4 = GetPrivateProfileInt("version", "digit4", 10, szIni);
         Digit5 = GetPrivateProfileInt("version", "digit5", 10, szIni);
         Digit6 = GetPrivateProfileInt("version", "digit6", 10, szIni);
         
         if ((10 == Digit4) || (10 == Digit5) || (10 == Digit6))
         {
            /* 3 digit version number */
            dDataVersion = (Digit1*100) + (Digit2*10) + Digit3;
         }         
         else
         {
            /* 6 digit version number */
            dDataVersion = (Digit1*100000) + (Digit2*10000) + (Digit3*1000) + (Digit4*100) + (Digit5*10) + Digit6;
         }   
      }         
            
      FindAllFiles(szInFolder);
      if (nFilelistCounter != 0) {
         CreateOutputImage();
      }
      else
      {
         printf("Error: Input folder \"%s\"not available\n", szInFolder);
      }
   }      
   
   Cleanup(0);
   
   return(0);
} /* main */

uLong zlib_adler32 (uLong adler, const Bytef *buf, uInt len)
{
   if (NULL == buf)
   {
      adler = ADLER_START_VALUE;
   }
   else
   {
      adler = adler32(adler, buf, len); 
   }
   
   return(adler);
}

/*** EOF ***/
